/*
 * Copyright (c) 2013-2014,2018 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Sat Jan 12 2013
 */
#include "acrbulkit.h"

#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <netinet/in.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


/* --- gobal data ---------------------------------------------------------- */

const char *port = "8869";
const char *dst = NULL;
int dotcp = 1;


/* --- usage --------------------------------------------------------------- */

/** Print usage on a channel.
 *
 * \param[in] channel   Output channel
 * \param[in] argv0     Program invocation name
 */

static void
usage(FILE *channel, char *argv0)
{
  fprintf(channel, "usage:  %s [-u] [-h] [-p port] host [message...]\n",
          basename(argv0));
}


/* --- setup_socket --------------------------------------------------------- */

static int
setup_socket(const char *host, const char *port)
{
  struct addrinfo hints;
  struct addrinfo *res, *res0;
  int s, e;

  /* lookup host addr */
  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_UNSPEC;
  if (dotcp) {
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
  } else {
    hints.ai_socktype = SOCK_DGRAM;
    hints.ai_protocol = IPPROTO_UDP;
  }
  hints.ai_flags = AI_CANONNAME;
  e = getaddrinfo(host, port, &hints, &res0);
  if (e) {
    warnx("%s: %s.%s", gai_strerror(e), host, port);
    return -1;
  }
  dst = strdup(res0->ai_canonname);

  /* socket */
  for(res = res0; res; res = res->ai_next) {
    s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (s < 0) continue;

    if (!connect(s, res->ai_addr, res->ai_addrlen)) {
      if (dotcp) {
        /* wait for acceptance */
        static const char response[] =
          "http/1.1 200 OK\r\n" "Content-Type: text/plain;\r\n" "\r\n";
        char c[sizeof(response)];

        do {
          e = recv(s, c, sizeof(response)-1, 0);
        } while( e == -1 && errno == EINTR);
        if (!strcasecmp(c, response)) break;
      } else break;
    }

    close(s);
    s = -1;
    errno = ECONNREFUSED;
  }
  if (s < 0) {
    warn("%s.%s", dst, port);
    return -1;
  }

  freeaddrinfo(res0);
  return s;
}


/* --- send_notify --------------------------------------------------------- */

static int
send_notify(int s, int argc, char * const argv[])
{
  char msg[65535] = "";
  char * const *p;
  ssize_t l;
  int i;

  for(i = argc, p = argv; i--; p++) {
    strncat(msg, *p, sizeof(msg) - 1 - strlen(msg));
    if (i > 0)
      strncat(msg, " ", sizeof(msg) - 1 - strlen(msg));
  }
  if (strlen(msg) >= sizeof(msg) - 1) {
    warnx("message too long");
    return 2;
  }

  l = send(s, msg, strlen(msg), 0);
  if (l != strlen(msg)) {
    warn("send");
    return 2;
  }

  return 0;
}


/* --- recv_print ---------------------------------------------------------- */

static int
print_recv(int s, int *status)
{
  static char msg[65535] = "";
  static ssize_t mlen;
  fd_set set;
  ssize_t l;
  int e;

  FD_ZERO(&set);
  FD_SET(s, &set);
  e = select(s + 1, &set, NULL, NULL, NULL);
  switch(e) {
    case -1: if (errno != EINTR) err(2, "select"); break;
    case 0: break;
    default:
      l = recv(s, msg, sizeof(msg), 0);
      if (l < 0) {
        if (errno == EAGAIN || errno == EWOULDBLOCK) break;
        err(2, "recv");
      }
      if (!l) {
        /* lookup status */
        *status = 0;
        if (msg[--mlen] != '\n') return 1;
        for(e = 1; isdigit((int)msg[--mlen]); e *= 10)
          *status += e * (msg[mlen] - '0');

        return 1;
      }

      mlen = l;
      if (write(STDOUT_FILENO, msg, l) != l)
        ;

      break;
  }

  return 0;
}


/* --- main ---------------------------------------------------------------- */

int
main(int argc, char *const argv[])
{
  extern char *optarg;
  extern int optind;
  char *argv0 = argv[0];

  int c, s;
  int status;

  /* parse command line options */
  while ((c = getopt(argc, argv, "hup:")) != -1) {
    switch (c) {
      case 'h':
        usage(stdout, argv0);
        exit(0);

      case 'u':
        dotcp = 0;
        break;

      case 'p':
        port = optarg;
        break;

      case '?':
      default:
        usage(stderr, argv0);
        exit(2);
    }
  }
  argc -= optind;
  argv += optind;
  if (argc < 1) {
    usage(stderr, argv0);
    exit(2);
  }

  /* setup socket */
  s = setup_socket(argv[0], port);
  if (s < 0)
    errx(2, "failed to notify %s.%s", dst?dst:argv[0], port);

  /* notify */
  if (send_notify(s, argc-1, argv+1))
    errx(2, "failed to notify %s.%s", dst?dst:argv[0], port);

  if (!dotcp) {
    close(s);
    printf("notified %s\n", dst);
    return 0;
  }

  /* wait for completion */
  shutdown(s, SHUT_WR);
  fcntl(s, F_SETFL, O_NONBLOCK);
  while(!print_recv(s, &status));
  close(s);

  return status;
}
